r"""
This module provides the core functionality for the computation of generic
semi-values. A **semi-value** is any valuation function with the form:

$$v_\text{semi}(i) = \sum_{i=1}^n w(k)
\sum_{S \subset D_{-i}^{(k)}} [U(S_{+i})-U(S)],$$

where the coefficients $w(k)$ satisfy the property:

$$\sum_{k=1}^n w(k) = 1.$$

??? Note
    For implementation consistency, we slightly depart from the common definition
    of semi-values, which includes a factor $1/n$ in the sum over subsets.
    Instead, we subsume this factor into the coefficient $w(k)$.

## Main components

The computation of a semi-value requires two components:

1. A **subset sampler** that generates subsets of the set $D$ of interest.
2. A **coefficient** $w(k)$ that assigns a weight to each subset size $k$.

Samplers can be found in [sampler][pydvl.value.sampler], and can be classified
into two categories: powerset samplers and permutation samplers. Powerset
samplers generate subsets of $D_{-i}$, while the permutation sampler generates
permutations of $D$. The former conform to the above definition of semi-values,
while the latter reformulates it as:

$$
v(i) = \frac{1}{n!} \sum_{\sigma \in \Pi(n)}
\tilde{w}( | \sigma_{:i} | )[U(\sigma_{:i} \cup \{i\}) − U(\sigma_{:i})],
$$

where $\sigma_{:i}$ denotes the set of indices in permutation sigma before the
position where $i$ appears (see [Data valuation][computing-data-values] for
details), and

$$ \tilde{w} (k) = n \binom{n - 1}{k} w (k) $$

is the weight correction due to the reformulation.

!!! Warning
    Both [PermutationSampler][pydvl.value.sampler.PermutationSampler] and
    [DeterministicPermutationSampler][pydvl.value.sampler.DeterministicPermutationSampler]
    require caching to be enabled or computation will be doubled wrt. a 'direct'
    implementation of permutation MC.

## Computing semi-values

Samplers and coefficients can be arbitrarily mixed by means of the main entry
point of this module,
[compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues].
There are several pre-defined coefficients, including the Shapley value of
(Ghorbani and Zou, 2019)[^1], the Banzhaf index of (Wang and Jia)[^3], and the Beta
coefficient of (Kwon and Zou, 2022)[^2]. For each of these methods, there is a
convenience wrapper function. Respectively, these are:
[compute_shapley_semivalues][pydvl.value.semivalues.compute_shapley_semivalues],
[compute_banzhaf_semivalues][pydvl.value.semivalues.compute_banzhaf_semivalues],
and [compute_beta_shapley_semivalues][pydvl.value.semivalues.compute_beta_shapley_semivalues].
instead.

!!! tip "Parallelization and batching"
    In order to ensure reproducibility and fine-grained control of
    parallelization, samples are generated in the main process and then
    distributed to worker processes for evaluation. For small sample sizes, this
    can lead to a significant overhead. To avoid this, we temporarily provide an
    additional argument `batch_size` to all methods which can improve
    performance with small models up to an order of magnitude. Note that this
    argument will be removed before version 1.0 in favour of a more general
    solution.


## References

[^1]: <a name="ghorbani_data_2019"></a>Ghorbani, A., Zou, J., 2019.
    [Data Shapley: Equitable Valuation of Data for Machine Learning](https://proceedings.mlr.press/v97/ghorbani19c.html).
    In: Proceedings of the 36th International Conference on Machine Learning, PMLR, pp. 2242–2251.

[^2]: <a name="kwon_beta_2022"></a>Kwon, Y. and Zou, J., 2022.
    [Beta Shapley: A Unified and Noise-reduced Data Valuation Framework for Machine Learning](https://arxiv.org/abs/2110.14049).
    In: Proceedings of the 25th International Conference on Artificial Intelligence and Statistics (AISTATS) 2022, Vol. 151. PMLR, Valencia, Spain.

[^3]: <a name="wang_data_2022"></a>Wang, J.T. and Jia, R., 2022.
    [Data Banzhaf: A Robust Data Valuation Framework for Machine Learning](https://arxiv.org/abs/2205.15466).
    ArXiv preprint arXiv:2205.15466.
"""
from __future__ import annotations

import logging
import math
import warnings
from enum import Enum
from itertools import islice
from typing import Iterable, List, Optional, Protocol, Tuple, Type, cast

import scipy as sp
from deprecate import deprecated
from tqdm import tqdm

from pydvl.parallel.config import ParallelConfig
from pydvl.utils import Utility
from pydvl.utils.types import IndexT, Seed
from pydvl.value import ValuationResult
from pydvl.value.sampler import (
    PermutationSampler,
    PowersetSampler,
    SampleT,
    StochasticSampler,
)
from pydvl.value.stopping import MaxUpdates, StoppingCriterion

__all__ = [
    "compute_banzhaf_semivalues",
    "compute_beta_shapley_semivalues",
    "compute_shapley_semivalues",
    "beta_coefficient",
    "banzhaf_coefficient",
    "shapley_coefficient",
    "compute_generic_semivalues",
    "compute_semivalues",
    "SemiValueMode",
]

log = logging.getLogger(__name__)


class SVCoefficient(Protocol):
    """The protocol that coefficients for the computation of semi-values must
    fulfill."""

    def __call__(self, n: int, k: int) -> float:
        """Computes the coefficient for a given subset size.

        Args:
            n: Total number of elements in the set.
            k: Size of the subset for which the coefficient is being computed
        """
        ...


MarginalT = Tuple[IndexT, float]


def _marginal(
    u: Utility, coefficient: SVCoefficient, samples: Iterable[SampleT]
) -> Tuple[MarginalT, ...]:
    """Computation of marginal utility. This is a helper function for
    [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues].

    Args:
        u: Utility object with model, data, and scoring function.
        coefficient: The semivalue coefficient and sampler weight
        samples: A collection of samples. Each sample is a tuple of index and subset of
            indices to compute a marginal utility.

    Returns:
        A collection of marginals. Each marginal is a tuple with index and its marginal
        utility.
    """
    n = len(u.data)
    marginals: List[MarginalT] = []
    for idx, s in samples:
        marginal = (u({idx}.union(s)) - u(s)) * coefficient(n, len(s))
        marginals.append((idx, marginal))
    return tuple(marginals)


# @deprecated(
#     target=compute_semivalues,  # TODO: rename this to compute_semivalues
#     deprecated_in="0.8.0",
#     remove_in="0.9.0",
# )
def compute_generic_semivalues(
    sampler: PowersetSampler[IndexT],
    u: Utility,
    coefficient: SVCoefficient,
    done: StoppingCriterion,
    *,
    batch_size: int = 1,
    skip_converged: bool = False,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    """Computes semi-values for a given utility function and subset sampler.

    Args:
        sampler: The subset sampler to use for utility computations.
        u: Utility object with model, data, and scoring function.
        coefficient: The semi-value coefficient
        done: Stopping criterion.
        batch_size: Number of marginal evaluations per single parallel job.
        skip_converged: Whether to skip marginal evaluations for indices that
            have already converged. **CAUTION**: This is only entirely safe if
            the stopping criterion is [MaxUpdates][pydvl.value.stopping.MaxUpdates].
            For any other stopping criterion, the convergence status of indices
            may change during the computation, or they may be marked as having
            converged even though in fact the estimated values are far from the
            true values (e.g. for
            [AbsoluteStandardError][pydvl.value.stopping.AbsoluteStandardError],
            you will probably have to carefully adjust the threshold).
        n_jobs: Number of parallel jobs to use.
        config: Object configuring parallel computation, with cluster
            address, number of cpus, etc.
        progress: Whether to display a progress bar.

    Returns:
        Object with the results.

    !!! warning "Deprecation notice"
        Parameter `batch_size` is for experimental use and will be removed in
        future versions.
    """
    from concurrent.futures import FIRST_COMPLETED, Future, wait

    from pydvl.parallel import effective_n_jobs, init_executor, init_parallel_backend

    if isinstance(sampler, PermutationSampler) and u.cache is None:
        log.warning(
            "PermutationSampler requires caching to be enabled or computation "
            "will be doubled wrt. a 'direct' implementation of permutation MC"
        )

    if batch_size != 1:
        warnings.warn(
            "Parameter `batch_size` is for experimental use and will be"
            " removed in future versions",
            DeprecationWarning,
        )

    result = ValuationResult.zeros(
        algorithm=f"semivalue-{str(sampler)}-{coefficient.__name__}",  # type: ignore
        indices=u.data.indices,
        data_names=u.data.data_names,
    )

    parallel_backend = init_parallel_backend(config)
    u = parallel_backend.put(u)
    correction = parallel_backend.put(
        lambda n, k: coefficient(n, k) * sampler.weight(n, k)
    )

    max_workers = effective_n_jobs(n_jobs, config)
    n_submitted_jobs = 2 * max_workers  # number of jobs in the queue

    sampler_it = iter(sampler)
    pbar = tqdm(disable=not progress, total=100, unit="%")

    with init_executor(
        max_workers=max_workers, config=config, cancel_futures=True
    ) as executor:
        pending: set[Future] = set()
        while True:
            pbar.n = 100 * done.completion()
            pbar.refresh()

            completed, pending = wait(pending, timeout=1, return_when=FIRST_COMPLETED)
            for future in completed:
                for idx, marginal in future.result():
                    result.update(idx, marginal)
                    if done(result):
                        return result

            # Ensure that we always have n_submitted_jobs running
            try:
                while len(pending) < n_submitted_jobs:
                    samples = tuple(islice(sampler_it, batch_size))
                    if len(samples) == 0:
                        raise StopIteration

                    # Filter out samples for indices that have already converged
                    filtered_samples = samples
                    if skip_converged and len(done.converged) > 0:
                        # TODO: cloudpickle can't pickle this on python 3.8:
                        # filtered_samples = filter(
                        #     lambda t: not done.converged[t[0]], samples
                        # )
                        filtered_samples = tuple(
                            (idx, sample)
                            for idx, sample in samples
                            if not done.converged[idx]
                        )

                    if filtered_samples:
                        pending.add(
                            executor.submit(
                                _marginal,
                                u=u,
                                coefficient=correction,
                                samples=filtered_samples,
                            )
                        )
            except StopIteration:
                if len(pending) == 0:
                    return result


def shapley_coefficient(n: int, k: int) -> float:
    return float(1 / math.comb(n - 1, k) / n)


def banzhaf_coefficient(n: int, k: int) -> float:
    return float(1 / 2 ** (n - 1))


def beta_coefficient(alpha: float, beta: float) -> SVCoefficient:
    # Leave this here in case we want to avoid depending on scipy
    # n = n or 1024
    # @lru_cache(maxsize=int(n * (n - 1) / 2))
    # def B(a: int, b: int) -> float:
    #     """Γ(a) * Γ(b) / Γ(a+b)"""
    #     ii = np.arange(1, b)
    #     return (ii / (a + ii)).prod() / a

    B = sp.special.beta
    const = B(alpha, beta)

    def beta_coefficient_w(n: int, k: int) -> float:
        """Beta coefficient"""
        j = k + 1
        w = B(j + beta - 1, n - j + alpha) / const
        # return math.comb(n - 1, j - 1) * w * n
        return float(w)

    return cast(SVCoefficient, beta_coefficient_w)


def compute_shapley_semivalues(
    u: Utility,
    *,
    done: StoppingCriterion = MaxUpdates(100),
    sampler_t: Type[StochasticSampler] = PermutationSampler,
    batch_size: int = 1,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> ValuationResult:
    """Computes Shapley values for a given utility function.

    This is a convenience wrapper for
    [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues]
    with the Shapley coefficient. Use
    [compute_shapley_values][pydvl.value.shapley.common.compute_shapley_values]
    for a more flexible interface and additional methods, including TMCS.

    Args:
        u: Utility object with model, data, and scoring function.
        done: Stopping criterion.
        sampler_t: The sampler type to use. See the
            [sampler][pydvl.value.sampler] module for a list.
        batch_size: Number of marginal evaluations per single parallel job.
        n_jobs: Number of parallel jobs to use.
        config: Object configuring parallel computation, with cluster
            address, number of cpus, etc.
        seed: Either an instance of a numpy random number generator or a seed
            for it.
        progress: Whether to display a progress bar.

    Returns:
        Object with the results.

    !!! warning "Deprecation notice"
        Parameter `batch_size` is for experimental use and will be removed in
        future versions.
    """
    # HACK: cannot infer return type because of useless IndexT, NameT
    return compute_generic_semivalues(  # type: ignore
        sampler_t(u.data.indices, seed=seed),
        u,
        shapley_coefficient,
        done,
        batch_size=batch_size,
        n_jobs=n_jobs,
        config=config,
        progress=progress,
    )


def compute_banzhaf_semivalues(
    u: Utility,
    *,
    done: StoppingCriterion = MaxUpdates(100),
    sampler_t: Type[StochasticSampler] = PermutationSampler,
    batch_size: int = 1,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> ValuationResult:
    """Computes Banzhaf values for a given utility function.

    This is a convenience wrapper for
    [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues]
    with the Banzhaf coefficient.

    Args:
        u: Utility object with model, data, and scoring function.
        done: Stopping criterion.
        sampler_t: The sampler type to use. See the
            [sampler][pydvl.value.sampler] module for a list.
        batch_size: Number of marginal evaluations per single parallel job.
        n_jobs: Number of parallel jobs to use.
        seed: Either an instance of a numpy random number generator or a seed
            for it.
        config: Object configuring parallel computation, with cluster address,
            number of cpus, etc.
        progress: Whether to display a progress bar.

    Returns:
        Object with the results.

    !!! warning "Deprecation notice"
        Parameter `batch_size` is for experimental use and will be removed in
        future versions.
    """
    # HACK: cannot infer return type because of useless IndexT, NameT
    return compute_generic_semivalues(  # type: ignore
        sampler_t(u.data.indices, seed=seed),
        u,
        banzhaf_coefficient,
        done,
        batch_size=batch_size,
        n_jobs=n_jobs,
        config=config,
        progress=progress,
    )


def compute_beta_shapley_semivalues(
    u: Utility,
    *,
    alpha: float = 1,
    beta: float = 1,
    done: StoppingCriterion = MaxUpdates(100),
    sampler_t: Type[StochasticSampler] = PermutationSampler,
    batch_size: int = 1,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> ValuationResult:
    """Computes Beta Shapley values for a given utility function.

    This is a convenience wrapper for
    [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues]
    with the Beta Shapley coefficient.

    Args:
        u: Utility object with model, data, and scoring function.
        alpha: Alpha parameter of the Beta distribution.
        beta: Beta parameter of the Beta distribution.
        done: Stopping criterion.
        sampler_t: The sampler type to use. See the
            [sampler][pydvl.value.sampler] module for a list.
        batch_size: Number of marginal evaluations per (parallelized) task.
        n_jobs: Number of parallel jobs to use.
        seed: Either an instance of a numpy random number generator or a seed for it.
        config: Object configuring parallel computation, with cluster address, number of
            cpus, etc.
        progress: Whether to display a progress bar.

    Returns:
        Object with the results.

    !!! warning "Deprecation notice"
        Parameter `batch_size` is for experimental use and will be removed in
        future versions.
    """
    # HACK: cannot infer return type because of useless IndexT, NameT
    return compute_generic_semivalues(  # type: ignore
        sampler_t(u.data.indices, seed=seed),
        u,
        beta_coefficient(alpha, beta),
        done,
        batch_size=batch_size,
        n_jobs=n_jobs,
        config=config,
        progress=progress,
    )


@deprecated(target=True, deprecated_in="0.7.0", remove_in="0.8.0")
class SemiValueMode(str, Enum):
    """Enumeration of semi-value modes.

    !!! warning "Deprecation notice"
        This enum and the associated methods are deprecated and will be removed
        in 0.8.0.
    """

    Shapley = "shapley"
    BetaShapley = "beta_shapley"
    Banzhaf = "banzhaf"


@deprecated(target=True, deprecated_in="0.7.0", remove_in="0.8.0")
def compute_semivalues(
    u: Utility,
    *,
    done: StoppingCriterion = MaxUpdates(100),
    mode: SemiValueMode = SemiValueMode.Shapley,
    sampler_t: Type[StochasticSampler] = PermutationSampler,
    batch_size: int = 1,
    n_jobs: int = 1,
    seed: Optional[Seed] = None,
    **kwargs,
) -> ValuationResult:
    """Convenience entry point for most common semi-value computations.

    !!! warning "Deprecation warning"
        This method is deprecated and will be replaced in 0.8.0 by the more
        general implementation of
        [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues].
        Use
        [compute_shapley_semivalues][pydvl.value.semivalues.compute_shapley_semivalues],
        [compute_banzhaf_semivalues][pydvl.value.semivalues.compute_banzhaf_semivalues],
        or
        [compute_beta_shapley_semivalues][pydvl.value.semivalues.compute_beta_shapley_semivalues]
        instead.

    The modes supported with this interface are the following. For greater
    flexibility use
    [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues]
    directly.

    - [SemiValueMode.Shapley][pydvl.value.semivalues.SemiValueMode]:
      Shapley values.
    - [SemiValueMode.BetaShapley][pydvl.value.semivalues.SemiValueMode.BetaShapley]:
      Implements the Beta Shapley semi-value as introduced in
      (Kwon and Zou, 2022)<sup><a href="#kwon_beta_2022">1</a></sup>.
      Pass additional keyword arguments `alpha` and `beta` to set the
      parameters of the Beta distribution (both default to 1).
    - [SemiValueMode.Banzhaf][SemiValueMode.Banzhaf]: Implements the Banzhaf
      semi-value as introduced in (Wang and Jia, 2022)<sup><a href="#wang_data_2022">1</a></sup>.

    See [[data-valuation]] for an overview of valuation.
    - [SemiValueMode.Banzhaf][pydvl.value.semivalues.SemiValueMode]: Implements
      the Banzhaf semi-value as introduced in [@wang_data_2022].

    Args:
        u: Utility object with model, data, and scoring function.
        done: Stopping criterion.
        mode: The semi-value mode to use. See
            [SemiValueMode][pydvl.value.semivalues.SemiValueMode] for a list.
        sampler_t: The sampler type to use. See [sampler][pydvl.value.sampler]
            for a list.
        batch_size: Number of marginal evaluations per (parallelized) task.
        n_jobs: Number of parallel jobs to use.
        seed: Either an instance of a numpy random number generator or a seed for it.
        kwargs: Additional keyword arguments passed to
            [compute_generic_semivalues][pydvl.value.semivalues.compute_generic_semivalues].

    Returns:
        Object with the results.

    !!! warning "Deprecation notice"
        Parameter `batch_size` is for experimental use and will be removed in
        future versions.
    """
    if mode == SemiValueMode.Shapley:
        coefficient = shapley_coefficient
    elif mode == SemiValueMode.BetaShapley:
        alpha = kwargs.pop("alpha", 1)
        beta = kwargs.pop("beta", 1)
        coefficient = beta_coefficient(alpha, beta)
    elif mode == SemiValueMode.Banzhaf:
        coefficient = banzhaf_coefficient
    else:
        raise ValueError(f"Unknown mode {mode}")
    coefficient = cast(SVCoefficient, coefficient)

    # HACK: cannot infer return type because of useless IndexT, NameT
    return compute_generic_semivalues(  # type: ignore
        sampler_t(u.data.indices, seed=seed),
        u,
        coefficient,
        done,
        n_jobs=n_jobs,
        batch_size=batch_size,
        **kwargs,
    )



from dataclasses import dataclass, field
from typing import Collection, Optional

__all__ = ["CachedFuncConfig"]


@dataclass
class CachedFuncConfig:
    """Configuration for cached functions and methods, providing
    memoization of function calls.

    Instances of this class are typically used as arguments for the construction
    of a [Utility][pydvl.utils.utility.Utility].

    Args:
        hash_prefix: Optional string prefix that be prepended to the cache key.
            This can be provided in order to guarantee cache reuse across runs.
        ignore_args: Do not take these keyword arguments into account when
            hashing the wrapped function for usage as key. This allows
            sharing the cache among different jobs for the same experiment run if
            the callable happens to have "nuisance" parameters like `job_id` which
            do not affect the result of the computation.
        time_threshold: Computations taking less time than this many seconds are
            not cached. A value of 0 means that it will always cache results.
        allow_repeated_evaluations: If `True`, repeated calls to a function
            with the same arguments will be allowed and outputs averaged until the
            running standard deviation of the mean stabilizes below
            `rtol_stderr * mean`.
        rtol_stderr: relative tolerance for repeated evaluations. More precisely,
            [memcached()][pydvl.utils.caching.memcached] will stop evaluating the function
            once the standard deviation of the mean is smaller than `rtol_stderr * mean`.
        min_repetitions: minimum number of times that a function evaluation
            on the same arguments is repeated before returning cached values. Useful
            for stochastic functions only. If the model training is very noisy, set
            this number to higher values to reduce variance.
    """

    hash_prefix: Optional[str] = None
    ignore_args: Collection[str] = field(default_factory=list)
    time_threshold: float = 0.3
    allow_repeated_evaluations: bool = False
    rtol_stderr: float = 0.1
    min_repetitions: int = 3



from deprecate import deprecated

from pydvl.utils import Utility
from pydvl.value.result import ValuationResult

from .loo import compute_loo

__all__ = ["naive_loo"]


@deprecated(
    target=compute_loo,
    deprecated_in="0.7.0",
    remove_in="0.8.0",
    args_extra=dict(n_jobs=1),
)
def naive_loo(u: Utility, *, progress: bool = True, **kwargs) -> ValuationResult:  # type: ignore
    """Deprecated. Use [compute_loo][pydvl.value.loo.compute_loo] instead."""
    pass



"""
This module provides classes and utilities for handling large arrays that are chunked
and lazily evaluated. It includes abstract base classes for converting between tensor
types and NumPy arrays, aggregating blocks of data, and abstract representations of
lazy arrays. Concrete implementations are provided for handling chunked lazy arrays
(chunked in one resp. two dimensions), with support for efficient storage and retrieval
using the Zarr library.
"""

from abc import ABC, abstractmethod
from typing import Callable, Generator, Generic, List, Optional, Tuple

import zarr
from numpy.typing import NDArray

from .base_influence_function_model import TensorType


class NumpyConverter(Generic[TensorType], ABC):
    """
    Base class for converting TensorType objects into numpy arrays and vice versa.
    """

    @abstractmethod
    def to_numpy(self, x: TensorType) -> NDArray:
        """Override this method for converting a TensorType object into a numpy array"""

    @abstractmethod
    def from_numpy(self, x: NDArray) -> TensorType:
        """Override this method for converting a numpy array into a TensorType object"""


class SequenceAggregator(Generic[TensorType], ABC):
    @abstractmethod
    def __call__(self, tensor_generator: Generator[TensorType, None, None]):
        """
        Aggregates tensors from a generator.

        Implement this method to define how a sequence of tensors, provided by a
        generator, should be combined.
        """


class ListAggregator(SequenceAggregator):
    def __call__(
        self, tensor_generator: Generator[TensorType, None, None]
    ) -> List[TensorType]:
        """
        Aggregates tensors from a single-level generator into a list. This method simply
        collects each tensor emitted by the generator into a single list.

        Args:
            tensor_generator: A generator that yields TensorType objects.

        Returns:
            A list containing all the tensors provided by the tensor_generator.
        """
        return [t for t in tensor_generator]


class NestedSequenceAggregator(Generic[TensorType], ABC):
    @abstractmethod
    def __call__(
        self,
        nested_generators_of_tensors: Generator[
            Generator[TensorType, None, None], None, None
        ],
    ):
        """
        Aggregates tensors from a generator of generators.

        Implement this method to specify how tensors, nested in two layers of
        generators, should be combined. Useful for complex data structures where tensors
        are not directly accessible in a flat list.
        """


class NestedListAggregator(NestedSequenceAggregator):
    def __call__(
        self,
        nested_generators_of_tensors: Generator[
            Generator[TensorType, None, None], None, None
        ],
    ) -> List[List[TensorType]]:
        """
         Aggregates tensors from a nested generator structure into a list of lists.
         Each inner generator is converted into a list of tensors, resulting in a nested
         list structure.

         Args:
             nested_generators_of_tensors: A generator of generators, where each inner
                generator yields TensorType objects.

        Returns:
            A list of lists, where each inner list contains tensors returned from one
                of the inner generators.
        """
        return [list(tensor_gen) for tensor_gen in nested_generators_of_tensors]


class LazyChunkSequence:
    """
    A class representing a chunked, and lazily evaluated array,
    where the chunking is restricted to the first dimension

    This class is designed to handle large arrays that don't fit in memory.
    It works by generating chunks of the array on demand and can
    also convert these chunks to a Zarr array
    for efficient storage and retrieval.

    Attributes:
        generator_factory: A factory function that returns
            a generator. This generator yields chunks of the large array when called.
    """

    def __init__(
        self, generator_factory: Callable[[], Generator[TensorType, None, None]]
    ):
        self.generator_factory = generator_factory

    def compute(self, aggregator: Optional[SequenceAggregator] = None):
        """
        Computes and optionally aggregates the chunks of the array using the provided
        aggregator. This method initiates the generation of chunks and then
        combines them according to the aggregator's logic.

        Args:
            aggregator: An optional aggregator for combining the chunks of
                the array. If None, a default ListAggregator is used to simply collect
                the chunks into a list.

        Returns:
            The aggregated result of all chunks of the array, the format of which
                depends on the aggregator used.

        """
        if aggregator is None:
            aggregator = ListAggregator()
        return aggregator(self.generator_factory())

    def to_zarr(
        self,
        path_or_url: str,
        converter: NumpyConverter,
        return_stored: bool = False,
        overwrite: bool = False,
    ) -> Optional[zarr.Array]:
        """
        Converts the array into Zarr format, a storage format optimized for large
        arrays, and stores it at the specified path or URL. This method is suitable for
        scenarios where the data needs to be saved for later use or for large datasets
        requiring efficient storage.

        Args:
            path_or_url: The file path or URL where the Zarr array will be stored.
            converter: A converter for transforming blocks into NumPy arrays
                compatible with Zarr.
            return_stored: If True, the method returns the stored Zarr array; otherwise,
                it returns None.
            overwrite: If True, overwrites existing data at the given path_or_url.
                If False, an error is raised in case of existing data.

        Returns:
            The Zarr array if return_stored is True; otherwise, None.
        """
        row_idx = 0
        z = None
        for block in self.generator_factory():
            numpy_block = converter.to_numpy(block)

            if z is None:
                z = self._initialize_zarr_array(numpy_block, path_or_url, overwrite)

            new_shape = self._new_shape_according_to_block(numpy_block, row_idx)
            z.resize(new_shape)

            z[row_idx : row_idx + numpy_block.shape[0]] = numpy_block
            row_idx += numpy_block.shape[0]

        return z if return_stored else None

    @staticmethod
    def _new_shape_according_to_block(
        block: NDArray,
        current_row_idx: int,
    ):
        return (current_row_idx + block.shape[0],) + block.shape[1:]

    @staticmethod
    def _initialize_zarr_array(block: NDArray, path_or_url: str, overwrite: bool):
        initial_shape = (0,) + block.shape[1:]
        return zarr.open(
            path_or_url,
            mode="w" if overwrite else "w-",
            shape=initial_shape,
            chunks=block.shape,
            dtype=block.dtype,
        )


class NestedLazyChunkSequence:
    """
    A class representing chunked, and lazily evaluated array, where the chunking is
    restricted to the first two dimensions.

    This class is designed for handling large arrays where individual chunks are
    loaded and processed lazily. It supports converting these chunks into a Zarr array
    for efficient storage and retrieval, with chunking applied along the first two
    dimensions.

    Attributes:
        generator_factory: A factory function that returns a generator of generators.
            Each inner generator yields chunks.
    """

    def __init__(
        self,
        generator_factory: Callable[
            [], Generator[Generator[TensorType, None, None], None, None]
        ],
    ):
        self.generator_factory = generator_factory

    def compute(self, aggregator: Optional[NestedSequenceAggregator] = None):
        """
        Computes and optionally aggregates the chunks of the array using the provided
        aggregator. This method initiates the generation of chunks and then
        combines them according to the aggregator's logic.

        Args:
            aggregator: An optional aggregator for combining the chunks of
                the array. If None, a default
                [NestedListAggregator][pydvl.influence.array.NestedListAggregator]
                is used to simply collect the chunks into a list of lists.

        Returns:
            The aggregated result of all chunks of the array, the format of which
            depends on the aggregator used.

        """
        if aggregator is None:
            aggregator = NestedListAggregator()
        return aggregator(self.generator_factory())

    def to_zarr(
        self,
        path_or_url: str,
        converter: NumpyConverter,
        return_stored: bool = False,
        overwrite: bool = False,
    ) -> Optional[zarr.Array]:
        """
        Converts the array into Zarr format, a storage format optimized for large
        arrays, and stores it at the specified path or URL. This method is suitable for
        scenarios where the data needs to be saved for later use or for large datasets
        requiring efficient storage.

        Args:
            path_or_url: The file path or URL where the Zarr array will be stored.
            converter: A converter for transforming blocks into NumPy arrays
                compatible with Zarr.
            return_stored: If True, the method returns the stored Zarr array;
                otherwise, it returns None.
            overwrite: If True, overwrites existing data at the given path_or_url.
                If False, an error is raised in case of existing data.

        Returns:
            The Zarr array if return_stored is True; otherwise, None.
        """

        row_idx = 0
        z = None
        numpy_block = None
        for row_blocks in self.generator_factory():
            col_idx = 0
            for block in row_blocks:
                numpy_block = converter.to_numpy(block)
                if z is None:
                    z = self._initialize_zarr_array(numpy_block, path_or_url, overwrite)
                new_shape = self._new_shape_according_to_block(
                    z, numpy_block, row_idx, col_idx
                )
                z.resize(new_shape)
                idx_slice_to_update = self._idx_slice_for_update(
                    numpy_block, row_idx, col_idx
                )
                z[idx_slice_to_update] = numpy_block

                col_idx += numpy_block.shape[1]

            if numpy_block is None:
                raise ValueError("Generator is empty")

            row_idx += numpy_block.shape[0]

        return z if return_stored else None

    @staticmethod
    def _idx_slice_for_update(
        block: NDArray, current_row_idx: int, current_col_idx
    ) -> Tuple[int, ...]:
        idx = [
            slice(current_row_idx, current_row_idx + block.shape[0]),
            slice(current_col_idx, current_col_idx + block.shape[1]),
        ]
        idx.extend(slice(s) for s in block.shape[2:])
        return tuple(idx)  # type:ignore

    @staticmethod
    def _new_shape_according_to_block(
        zarr_array: zarr.Array,
        block: NDArray,
        current_row_idx: int,
        current_col_idx: int,
    ):
        return (
            max(current_row_idx + block.shape[0], zarr_array.shape[0]),
            max(current_col_idx + block.shape[1], zarr_array.shape[1]),
        ) + block.shape[2:]

    @staticmethod
    def _initialize_zarr_array(
        block: NDArray, path_or_url: str, overwrite: bool
    ) -> zarr.Array:
        fixed_shape = block.shape[2:]
        initial_shape = (0, 0) + fixed_shape
        chunk_size = block.shape[:2] + fixed_shape
        return zarr.open(
            path_or_url,
            mode="w" if overwrite else "w-",
            shape=initial_shape,
            chunks=chunk_size,
            dtype=block.dtype,
        )



"""
This module provides a common interface to parallelization backends. The list of
supported backends is [here][pydvl.parallel.backends]. Backends can be
selected with the `backend` argument of an instance of
[ParallelConfig][pydvl.utils.config.ParallelConfig], as seen in the examples
below.

We use [executors][concurrent.futures.Executor] to submit tasks in parallel. The
basic high-level pattern is

```python
from pydvl.parallel import init_executor, ParallelConfig

config = ParallelConfig(backend="ray")
with init_executor(max_workers=1, config=config) as executor:
    future = executor.submit(lambda x: x + 1, 1)
    result = future.result()
assert result == 2
```

Running a map-reduce job is also easy:

```python
from pydvl.parallel import init_executor, ParallelConfig

config = ParallelConfig(backend="joblib")
with init_executor(config=config) as executor:
    results = list(executor.map(lambda x: x + 1, range(5)))
assert results == [1, 2, 3, 4, 5]
```

There is an alternative map-reduce implementation
[MapReduceJob][pydvl.parallel.map_reduce.MapReduceJob] which internally
uses joblib's higher level API with `Parallel()`
"""
# HACK to avoid circular imports
from ..utils.types import *  # pylint: disable=wrong-import-order
from .backend import *
from .backends import *
from .config import *
from .futures import *
from .map_reduce import *

if len(BaseParallelBackend.BACKENDS) == 0:
    raise ImportError("No parallel backend found. Please install ray or joblib.")



import logging
import warnings
from typing import Optional

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from pydvl.utils import Utility, powerset
from pydvl.value.least_core.common import LeastCoreProblem, lc_solve_problem
from pydvl.value.result import ValuationResult

__all__ = ["exact_least_core", "lc_prepare_problem"]

logger = logging.getLogger(__name__)


def exact_least_core(
    u: Utility,
    *,
    non_negative_subsidy: bool = False,
    solver_options: Optional[dict] = None,
    options: Optional[dict] = None,
    progress: bool = True,
) -> ValuationResult:
    r"""Computes the exact Least Core values.

    !!! Note
        If the training set contains more than 20 instances a warning is printed
        because the computation is very expensive. This method is mostly used for
        internal testing and simple use cases. Please refer to the
        [Monte Carlo method][pydvl.value.least_core.montecarlo.montecarlo_least_core]
        for practical applications.

    The least core is the solution to the following Linear Programming problem:

    $$
    \begin{array}{lll}
    \text{minimize} & \displaystyle{e} & \\
    \text{subject to} & \displaystyle\sum_{i\in N} x_{i} = v(N) & \\
    & \displaystyle\sum_{i\in S} x_{i} + e \geq v(S) &, \forall S \subseteq N \\
    \end{array}
    $$

    Where $N = \{1, 2, \dots, n\}$ are the training set's indices.

    Args:
        u: Utility object with model, data, and scoring function
            non_negative_subsidy: If True, the least core subsidy $e$ is constrained
            to be non-negative.
        solver_options: Dictionary of options that will be used to select a solver
            and to configure it. Refer to the [cvxpy's
            documentation](https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options)
            for all possible options.
        options: (Deprecated) Dictionary of solver options. Use `solver_options`
            instead.
        progress: If True, shows a tqdm progress bar

    Returns:
        Object with the data values and the least core value.
    """
    n = len(u.data)
    if n > 20:  # Arbitrary choice, will depend on time required, caching, etc.
        warnings.warn(f"Large dataset! Computation requires 2^{n} calls to model.fit()")

    # TODO: remove this before releasing version 0.7.0
    if options:
        warnings.warn(
            DeprecationWarning(
                "Passing solver options as kwargs was deprecated in "
                "0.6.0, will "
                "be removed in 0.7.0. `Use solver_options` instead."
            )
        )
        if solver_options is None:
            solver_options = options
        else:
            solver_options.update(options)

    problem = lc_prepare_problem(u, progress=progress)
    return lc_solve_problem(
        problem=problem,
        u=u,
        algorithm="exact_least_core",
        non_negative_subsidy=non_negative_subsidy,
        solver_options=solver_options,
    )


def lc_prepare_problem(u: Utility, progress: bool = False) -> LeastCoreProblem:
    """Prepares a linear problem with all subsets of the data
    Use this to separate the problem preparation from the solving with
    [lc_solve_problem()][pydvl.value.least_core.common.lc_solve_problem]. Useful for
    parallel execution of multiple experiments.

    See [exact_least_core()][pydvl.value.least_core.naive.exact_least_core] for argument
    descriptions.
    """
    n = len(u.data)

    logger.debug("Building vectors and matrices for linear programming problem")
    powerset_size = 2**n
    A_lb = np.zeros((powerset_size, n))

    logger.debug("Iterating over all subsets")
    utility_values = np.zeros(powerset_size)
    for i, subset in enumerate(  # type: ignore
        tqdm(
            powerset(u.data.indices),
            disable=not progress,
            total=powerset_size - 1,
            position=0,
        )
    ):
        indices: NDArray[np.bool_] = np.zeros(n, dtype=bool)
        indices[list(subset)] = True
        A_lb[i, indices] = 1
        utility_values[i] = u(subset)  # type: ignore

    return LeastCoreProblem(utility_values, A_lb)



"""
This package contains algorithms for the computation of the influence function.

> **Warning:** Much of the code in this package is experimental or untested and is
subject to modification. In particular, the package structure and basic API will
probably change.

"""
from .base_influence_function_model import InfluenceMode
from .influence_calculator import (
    DaskInfluenceCalculator,
    DisableClientSingleThreadCheck,
    SequentialInfluenceCalculator,
)



import logging
import warnings
from typing import Iterable, Optional

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from pydvl.parallel import MapReduceJob, ParallelConfig, effective_n_jobs
from pydvl.utils.numeric import random_powerset
from pydvl.utils.types import Seed
from pydvl.utils.utility import Utility
from pydvl.value.least_core.common import LeastCoreProblem, lc_solve_problem
from pydvl.value.result import ValuationResult

logger = logging.getLogger(__name__)


__all__ = ["montecarlo_least_core", "mclc_prepare_problem"]


def montecarlo_least_core(
    u: Utility,
    n_iterations: int,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    non_negative_subsidy: bool = False,
    solver_options: Optional[dict] = None,
    options: Optional[dict] = None,
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> ValuationResult:
    r"""Computes approximate Least Core values using a Monte Carlo approach.

    $$
    \begin{array}{lll}
    \text{minimize} & \displaystyle{e} & \\
    \text{subject to} & \displaystyle\sum_{i\in N} x_{i} = v(N) & \\
    & \displaystyle\sum_{i\in S} x_{i} + e \geq v(S) & ,
    \forall S \in \{S_1, S_2, \dots, S_m \overset{\mathrm{iid}}{\sim} U(2^N) \}
    \end{array}
    $$

    Where:

    * $U(2^N)$ is the uniform distribution over the powerset of $N$.
    * $m$ is the number of subsets that will be sampled and whose utility will
      be computed and used to compute the data values.

    Args:
        u: Utility object with model, data, and scoring function
        n_iterations: total number of iterations to use
        n_jobs: number of jobs across which to distribute the computation
        config: Object configuring parallel computation, with cluster
            address, number of cpus, etc.
        non_negative_subsidy: If True, the least core subsidy $e$ is constrained
            to be non-negative.
        solver_options: Dictionary of options that will be used to select a solver
            and to configure it. Refer to [cvxpy's
            documentation](https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options)
            for all possible options.
        options: (Deprecated) Dictionary of solver options. Use solver_options
            instead.
        progress: If True, shows a tqdm progress bar
        seed: Either an instance of a numpy random number generator or a seed for it.

    Returns:
        Object with the data values and the least core value.
    """
    # TODO: remove this before releasing version 0.7.0
    if options:
        warnings.warn(
            DeprecationWarning(
                "Passing solver options as kwargs was deprecated in "
                "0.6.0, will be removed in 0.7.0. `Use solver_options` "
                "instead."
            )
        )
        if solver_options is None:
            solver_options = options
        else:
            solver_options.update(options)

    problem = mclc_prepare_problem(
        u, n_iterations, n_jobs=n_jobs, config=config, progress=progress, seed=seed
    )
    return lc_solve_problem(
        problem,
        u=u,
        algorithm="montecarlo_least_core",
        non_negative_subsidy=non_negative_subsidy,
        solver_options=solver_options,
    )


def mclc_prepare_problem(
    u: Utility,
    n_iterations: int,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> LeastCoreProblem:
    """Prepares a linear problem by sampling subsets of the data. Use this to
    separate the problem preparation from the solving with
    [lc_solve_problem()][pydvl.value.least_core.common.lc_solve_problem]. Useful
    for parallel execution of multiple experiments.

    See
    [montecarlo_least_core][pydvl.value.least_core.montecarlo.montecarlo_least_core]
    for argument descriptions.
    """
    n = len(u.data)

    if n_iterations < n:
        warnings.warn(
            f"Number of iterations '{n_iterations}' is smaller the size of the dataset '{n}'. "
            f"This is not optimal because in the worst case we need at least '{n}' constraints "
            "to satisfy the individual rationality condition."
        )

    if n_iterations > 2**n:
        warnings.warn(
            f"Passed n_iterations is greater than the number subsets! "
            f"Setting it to 2^{n}",
            RuntimeWarning,
        )
        n_iterations = 2**n

    iterations_per_job = max(1, n_iterations // effective_n_jobs(n_jobs, config))

    map_reduce_job: MapReduceJob["Utility", "LeastCoreProblem"] = MapReduceJob(
        inputs=u,
        map_func=_montecarlo_least_core,
        reduce_func=_reduce_func,
        map_kwargs=dict(n_iterations=iterations_per_job, progress=progress),
        n_jobs=n_jobs,
        config=config,
    )

    return map_reduce_job(seed=seed)


def _montecarlo_least_core(
    u: Utility,
    n_iterations: int,
    *,
    progress: bool = False,
    job_id: int = 1,
    seed: Optional[Seed] = None,
) -> LeastCoreProblem:
    """Computes utility values and the Least Core upper bound matrix for a given
    number of iterations.

    Args:
        u: Utility object with model, data, and scoring function
        n_iterations: total number of iterations to use
        progress: If True, shows a tqdm progress bar
        job_id: Integer id used to determine the position of the progress bar
        seed: Either an instance of a numpy random number generator or a seed for it.

    Returns:
        A solution
    """
    n = len(u.data)

    utility_values = np.zeros(n_iterations)

    # Randomly sample subsets of full dataset
    rng = np.random.default_rng(seed)
    power_set = random_powerset(u.data.indices, n_samples=n_iterations, seed=rng)

    A_lb = np.zeros((n_iterations, n))

    for i, subset in enumerate(
        tqdm(power_set, disable=not progress, total=n_iterations, position=job_id)
    ):
        indices: NDArray[np.bool_] = np.zeros(n, dtype=bool)
        indices[list(subset)] = True
        A_lb[i, indices] = 1
        utility_values[i] = u(subset)

    return LeastCoreProblem(utility_values, A_lb)


def _reduce_func(results: Iterable[LeastCoreProblem]) -> LeastCoreProblem:
    """Combines the results from different parallel runs of
    [_montecarlo_least_core()][pydvl.value.least_core.montecarlo._montecarlo_least_core]
    """
    utility_values_list, A_lb_list = zip(*results)
    utility_values = np.concatenate(utility_values_list)
    A_lb = np.concatenate(A_lb_list)
    return LeastCoreProblem(utility_values, A_lb)



import logging
import queue
import sys
import threading
import time
import types
from concurrent.futures import Executor, Future
from typing import Any, Callable, Optional, TypeVar
from weakref import WeakSet, ref

import ray
from deprecate import deprecated

from pydvl.parallel.config import ParallelConfig

__all__ = ["RayExecutor"]

from pydvl.parallel import CancellationPolicy

T = TypeVar("T")

logger = logging.getLogger(__name__)


class RayExecutor(Executor):
    """Asynchronous executor using Ray that implements the concurrent.futures API.

    It shouldn't be initialized directly. You should instead call
    [init_executor()][pydvl.parallel.futures.init_executor].

    Args:
        max_workers: Maximum number of concurrent tasks. Each task can request
            itself any number of vCPUs. You must ensure the product of this
            value and the n_cpus_per_job parameter passed to submit() does not
            exceed available cluster resources. If set to `None`, it will
            default to the total number of vCPUs in the ray cluster.
        config: instance of [ParallelConfig][pydvl.utils.config.ParallelConfig]
            with cluster address, number of cpus, etc.
        cancel_futures: Select which futures will be cancelled when exiting this
            context manager. `Pending` is the default, which will cancel all
            pending futures, but not running ones, as done by
            [concurrent.futures.ProcessPoolExecutor][]. Additionally, `All`
            cancels all pending and running futures, and `None` doesn't cancel
            any. See [CancellationPolicy][pydvl.parallel.backend.CancellationPolicy]
    """

    @deprecated(
        target=True,
        deprecated_in="0.7.0",
        remove_in="0.8.0",
        args_mapping={"cancel_futures_on_exit": "cancel_futures"},
    )
    def __init__(
        self,
        max_workers: Optional[int] = None,
        *,
        config: ParallelConfig = ParallelConfig(),
        cancel_futures: CancellationPolicy = CancellationPolicy.ALL,
    ):
        if config.backend != "ray":
            raise ValueError(
                f"Parallel backend must be set to 'ray' and not '{config.backend}'"
            )
        if max_workers is not None:
            if max_workers <= 0:
                raise ValueError("max_workers must be greater than 0")
            max_workers = max_workers

        if isinstance(cancel_futures, CancellationPolicy):
            self._cancel_futures = cancel_futures
        else:
            self._cancel_futures = (
                CancellationPolicy.PENDING
                if cancel_futures
                else CancellationPolicy.NONE
            )

        self.config = {"address": config.address, "logging_level": config.logging_level}
        if config.address is None:
            self.config["num_cpus"] = config.n_cpus_local

        if not ray.is_initialized():
            ray.init(**self.config)

        self._max_workers = max_workers
        if self._max_workers is None:
            self._max_workers = int(ray._private.state.cluster_resources()["CPU"])

        self._shutdown = False
        self._shutdown_lock = threading.Lock()
        self._queue_lock = threading.Lock()
        self._work_queue: "queue.Queue[Optional[_WorkItem]]" = queue.Queue(
            maxsize=self._max_workers
        )
        self._pending_queue: "queue.SimpleQueue[Optional[_WorkItem]]" = (
            queue.SimpleQueue()
        )

        # Work Item Manager Thread
        self._work_item_manager_thread: Optional[_WorkItemManagerThread] = None

    def submit(self, fn: Callable[..., T], *args, **kwargs) -> "Future[T]":
        r"""Submits a callable to be executed with the given arguments.

        Schedules the callable to be executed as fn(\*args, \**kwargs)
        and returns a Future instance representing the execution of the callable.

        Args:
            fn: Callable.
            args: Positional arguments that will be passed to `fn`.
            kwargs: Keyword arguments that will be passed to `fn`.
                It can also optionally contain options for the ray remote function
                as a dictionary as the keyword argument `remote_function_options`.
        Returns:
            A Future representing the given call.

        Raises:
            RuntimeError: If a task is submitted after the executor has been shut down.
        """
        with self._shutdown_lock:
            logger.debug("executor acquired shutdown lock")
            if self._shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")

            logging.debug("Creating future and putting work item in work queue")
            future: "Future[T]" = Future()
            remote_function_options = kwargs.pop("remote_function_options", None)
            w = _WorkItem(
                future,
                fn,
                args,
                kwargs,
                remote_function_options=remote_function_options,
            )
            self._put_work_item_in_queue(w)
            # We delay starting the thread until the first call to submit
            self._start_work_item_manager_thread()
            return future

    def shutdown(
        self, wait: bool = True, *, cancel_futures: Optional[bool] = None
    ) -> None:
        """Clean up the resources associated with the Executor.

        This method tries to mimic the behaviour of
        [Executor.shutdown][concurrent.futures.Executor.shutdown]
        while allowing one more value for ``cancel_futures`` which instructs it
        to use the [CancellationPolicy][pydvl.parallel.backend.CancellationPolicy]
        defined upon construction.

        Args:
            wait: Whether to wait for pending futures to finish.
            cancel_futures: Overrides the executor's default policy for
                cancelling futures on exit. If ``True``, all pending futures are
                cancelled, and if ``False``, no futures are cancelled. If ``None``
                (default), the executor's policy set at initialization is used.
        """
        logger.debug("executor shutting down")
        with self._shutdown_lock:
            logger.debug("executor acquired shutdown lock")
            self._shutdown = True
            self._cancel_futures = {
                None: self._cancel_futures,
                True: CancellationPolicy.PENDING,
                False: CancellationPolicy.NONE,
            }[cancel_futures]

        if wait:
            logger.debug("executor waiting for futures to finish")
            if self._work_item_manager_thread is not None:
                # Putting None in the queue to signal
                # to work item manager thread that we are shutting down
                self._put_work_item_in_queue(None)
                logger.debug(
                    "executor waiting for work item manager thread to terminate"
                )
                self._work_item_manager_thread.join()
            # To reduce the risk of opening too many files, remove references to
            # objects that use file descriptors.
            self._work_item_manager_thread = None
            del self._work_queue
            del self._pending_queue

    def _put_work_item_in_queue(self, work_item: Optional["_WorkItem"]) -> None:
        with self._queue_lock:
            logger.debug("executor acquired queue lock")
            try:
                self._work_queue.put_nowait(work_item)
            except queue.Full:
                self._pending_queue.put_nowait(work_item)

    def _start_work_item_manager_thread(self) -> None:
        if self._work_item_manager_thread is None:
            self._work_item_manager_thread = _WorkItemManagerThread(self)
            self._work_item_manager_thread.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context related to the RayExecutor object."""
        self.shutdown()
        return False


class _WorkItem:
    """Inspired by code from: [concurrent.futures.thread][]"""

    def __init__(
        self,
        future: Future,
        fn: Callable,
        args: Any,
        kwargs: Any,
        *,
        remote_function_options: Optional[dict] = None,
    ):
        self.future = future
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.remote_function_options = remote_function_options or {"num_cpus": 1.0}

    def run(self) -> None:
        if not self.future.set_running_or_notify_cancel():
            return

        remote_fn = ray.remote(self.fn)
        ref = remote_fn.options(
            name=self.fn.__name__, **self.remote_function_options
        ).remote(*self.args, **self.kwargs)

        # Almost verbatim copy of `future()` method of ClientObjectRef
        def set_future(data: Any) -> None:
            """Schedules a callback to set the exception or result
            in the Future."""

            if isinstance(data, Exception):
                self.future.set_exception(data)
            else:
                self.future.set_result(data)

        ref._on_completed(set_future)
        # Prevent this object ref from being released.
        self.future.object_ref = ref  # type: ignore

    if sys.version_info >= (3, 9):
        __class_getitem__ = classmethod(types.GenericAlias)


class _WorkItemManagerThread(threading.Thread):
    """Manages submitting the work items and throttling.

    It runs in a local thread.
    Args:
        executor: An instance of RayExecutor that owns
        this thread. A weakref will be owned by the manager as well as
        references to internal objects used to introspect the state of
        the executor.
    """

    def __init__(self, executor: RayExecutor):
        self.executor_reference = ref(executor)
        self.shutdown_lock: threading.Lock = executor._shutdown_lock
        self.queue_lock: threading.Lock = executor._queue_lock
        self.work_queue: "queue.Queue[Optional[_WorkItem]]" = executor._work_queue
        self.pending_queue: "queue.SimpleQueue[Optional[_WorkItem]]" = (
            executor._pending_queue
        )
        self.submitted_futures: "WeakSet[Future]" = WeakSet()
        super().__init__()

    def run(self) -> None:
        logger.debug("starting work item manager thread main loop")
        while True:
            # This is used to avoid using too much CPU
            time.sleep(0.01)
            try:
                self.add_pending_item_to_work_queue()
                self.submit_work_item()
            finally:
                if self.is_shutting_down():
                    self.flag_executor_shutting_down()

                    # Since no new work items can be added,
                    # it is safe to shut down this thread
                    # if there are no more work items.
                    if self.work_queue.empty() and self.pending_queue.empty():
                        break
        logger.debug("exiting work item manager thread main loop")

    def add_pending_item_to_work_queue(self) -> None:
        # Fills work_queue with _WorkItems from pending_queue.
        # This function never blocks.
        while True:
            # This is used to avoid using too much CPU
            time.sleep(0.01)
            with self.queue_lock:
                logger.debug("work item manager thread acquired queue lock")
                # If the work queue is not full,
                # Move a work item from the pending queue, if not empty,
                # to the work queue
                if self.work_queue.full():
                    return
                try:
                    work_item = self.pending_queue.get_nowait()
                except queue.Empty:
                    logger.debug("pending queue is empty")
                    return
                else:
                    if work_item is None:
                        self.work_queue.put_nowait(work_item)
                        return
                    logger.debug("moving work item from pending queue to work queue")
                    self.work_queue.put(work_item)
                    del work_item

    def submit_work_item(self) -> None:
        with self.queue_lock:
            logger.debug("work item manager thread acquired queue lock")
            # Try to get a work item
            # If it is None, we break from the loop.
            # Otherwise, we submit the future.
            try:
                logger.debug("getting work item from work queue")
                work_item = self.work_queue.get_nowait()
                if work_item is None:
                    return
            except queue.Empty:
                logger.debug("work queue is empty")
                return
        logger.debug("Submitting work item")
        work_item.run()
        self.submitted_futures.add(work_item.future)
        # Delete references to object
        del work_item

    def is_shutting_down(self) -> bool:
        # Check whether we should start shutting down the executor.
        executor = self.executor_reference()
        # No more work items can be added if:
        #   - The executor that owns this worker has been collected OR
        #   - The executor that owns this worker has been shutdown.
        with self.shutdown_lock:
            logger.debug("work item manager thread acquired shutdown lock")
            if executor is None or executor._shutdown:
                # Flag the executor as shutting down as early as possible if it
                # is not gc-ed yet.
                if executor is not None:
                    executor._shutdown = True
                return True
            return False

    def flag_executor_shutting_down(self):
        # Flag the executor as shutting down and cancel remaining tasks if
        # requested as early as possible if it is not gc-ed yet.
        executor = self.executor_reference()
        with self.shutdown_lock:
            logger.debug("work item manager thread acquired shutdown lock")
            if executor is None:
                return
            executor._shutdown = True

            if executor._cancel_futures & CancellationPolicy.PENDING:
                # Drain all work items from the queues,
                # and then cancel their associated futures.
                # We empty the pending queue first.
                logger.debug("cancelling pending work items")
                while True:
                    with self.queue_lock:
                        try:
                            work_item = self.pending_queue.get_nowait()
                        except queue.Empty:
                            break
                        if work_item is not None:
                            work_item.future.cancel()
                            del work_item
                while True:
                    with self.queue_lock:
                        try:
                            work_item = self.work_queue.get_nowait()
                        except queue.Empty:
                            break
                        if work_item is not None:
                            work_item.future.cancel()
                            del work_item
                # Make sure we do this only once to not waste time looping
                # on running processes over and over.
                executor._cancel_futures &= ~CancellationPolicy.PENDING

            if executor._cancel_futures & CancellationPolicy.RUNNING:
                logger.debug("forcefully cancelling running futures")
                # We cancel the future's object references
                # We cannot cancel a running future object.
                for future in self.submitted_futures:
                    ray.cancel(future.object_ref)  # type: ignore
                # Make sure we do this only once to not waste time looping
                # on running processes over and over.
                executor._cancel_futures &= ~CancellationPolicy.RUNNING
